@@ -27,6 +27,7 @@ through the code to commit the whitespace formatting.
 """
 
 from yapf.yapflib import format_token
+from yapf.yapflib import object_state
 from yapf.yapflib import split_penalty
 from yapf.yapflib import style
 from yapf.yapflib import unwrapped_line
@@ -49,6 +50,8 @@ class FormatDecisionState(object):
     previous: The previous format decision state in the decision tree.
     stack: A stack (of _ParenState) keeping track of properties applying to
       parenthesis levels.
+    comp_stack: A stack (of ComprehensionState) keeping track of properties
+      applying to comprehensions.
     ignore_stack_for_comparison: Ignore the stack of _ParenState for state
       comparison.
   """
@@ -71,6 +74,7 @@ class FormatDecisionState(object):
     self.lowest_level_on_line = 0
     self.ignore_stack_for_comparison = False
     self.stack = [_ParenState(first_indent, first_indent)]
+    self.comp_stack = []
     self.first_indent = first_indent
     self.newline = False
     self.previous = None
@@ -90,6 +94,7 @@ class FormatDecisionState(object):
     new.newline = self.newline
     new.previous = self.previous
     new.stack = [state.Clone() for state in self.stack]
+    new.comp_stack = [state.Clone() for state in self.comp_stack]
     return new
 
   def __eq__(self, other):
@@ -102,7 +107,8 @@ class FormatDecisionState(object):
             self.start_of_line_level == other.start_of_line_level and
             self.lowest_level_on_line == other.lowest_level_on_line and
             (self.ignore_stack_for_comparison or
-             other.ignore_stack_for_comparison or self.stack == other.stack))
+             other.ignore_stack_for_comparison or
+             self.stack == other.stack and self.comp_stack == other.comp_stack))
 
   def __ne__(self, other):
     return not self == other
@@ -256,21 +262,25 @@ class FormatDecisionState(object):
       #
       # or when a string formatting syntax.
       func_call_or_string_format = False
+      tok = current.next_token
       if current.is_name:
-        tok = current.next_token
         while tok and (tok.is_name or tok.value == '.'):
           tok = tok.next_token
         func_call_or_string_format = tok and tok.value == '('
       elif current.is_string:
-        tok = current.next_token
         while tok and tok.is_string:
           tok = tok.next_token
         func_call_or_string_format = tok and tok.value == '%'
       if func_call_or_string_format:
         open_bracket = unwrapped_line.IsSurroundedByBrackets(current)
-        if open_bracket and open_bracket.value in '[{':
-          if not self._FitsOnLine(open_bracket, open_bracket.matching_bracket):
-            return True
+        if open_bracket:
+          if open_bracket.value in '[{':
+            if not self._FitsOnLine(open_bracket,
+                                    open_bracket.matching_bracket):
+              return True
+          elif tok.value == '(':
+            if not self._FitsOnLine(current, tok.matching_bracket):
+              return True
 
     ###########################################################################
     # Dict/Set Splitting
@@ -352,6 +362,10 @@ class FormatDecisionState(object):
               self.stack[-1].indent)
           return arglist_length > self.column_limit
 
+    if (current.value not in '{)' and previous.value == '(' and
+        self._ArgumentListHasDictionaryEntry(current)):
+      return True
+
     if style.Get('SPLIT_ARGUMENTS_WHEN_COMMA_TERMINATED'):
       # Split before arguments in a function call or definition if the
       # arguments are terminated by a comma.
@@ -371,18 +385,21 @@ class FormatDecisionState(object):
           (opening.previous_token.is_name or
            opening.previous_token.value in {'*', '**'})):
         is_func_call = False
-        token = current
-        while token:
-          if token.value == '(':
+        opening = current
+        while opening:
+          if opening.value == '(':
             is_func_call = True
             break
-          if (not (token.is_name or token.value in {'*', '**'}) and
-              token.value != '.'):
+          if (not (opening.is_name or opening.value in {'*', '**'}) and
+              opening.value != '.'):
             break
-          token = token.next_token
+          opening = opening.next_token
 
         if is_func_call:
-          if not self._FitsOnLine(current, opening.matching_bracket):
+          if (not self._FitsOnLine(current, opening.matching_bracket) or
+              (opening.matching_bracket.next_token and
+               opening.matching_bracket.next_token.value != ',' and
+               not opening.matching_bracket.next_token.ClosesScope())):
             return True
 
     pprevious = previous.previous_token
@@ -404,16 +421,15 @@ class FormatDecisionState(object):
         return True
 
     if (previous.OpensScope() and not current.OpensScope() and
+        not current.is_comment and
         format_token.Subtype.SUBSCRIPT_BRACKET not in previous.subtypes):
-      if not current.is_comment:
-        if pprevious and not pprevious.is_keyword and not pprevious.is_name:
-          # We want to split if there's a comment in the container.
-          token = current
-          while token != previous.matching_bracket:
-            if token.is_comment:
-              return True
-            token = token.next_token
-
+      if pprevious and not pprevious.is_keyword and not pprevious.is_name:
+        # We want to split if there's a comment in the container.
+        token = current
+        while token != previous.matching_bracket:
+          if token.is_comment:
+            return True
+          token = token.next_token
       if previous.value == '(':
         pptoken = previous.previous_token
         if not pptoken or not pptoken.is_name:
@@ -427,27 +443,10 @@ class FormatDecisionState(object):
             return current.next_token != previous.matching_bracket
       else:
         # Split after the opening of a container if it doesn't fit on the
-        # current line or if it has a comment.
+        # current line.
         if not self._FitsOnLine(previous, previous.matching_bracket):
           return True
 
-    ###########################################################################
-    # List Comprehension Splitting
-    if (format_token.Subtype.COMP_FOR in current.subtypes and
-        format_token.Subtype.COMP_FOR not in previous.subtypes):
-      # Split at the beginning of a list comprehension.
-      length = _GetLengthOfSubtype(current, format_token.Subtype.COMP_FOR,
-                                   format_token.Subtype.COMP_IF)
-      if length + self.column > self.column_limit:
-        return True
-
-    if (format_token.Subtype.COMP_IF in current.subtypes and
-        format_token.Subtype.COMP_IF not in previous.subtypes):
-      # Split at the beginning of an if expression.
-      length = _GetLengthOfSubtype(current, format_token.Subtype.COMP_IF)
-      if length + self.column > self.column_limit:
-        return True
-
     ###########################################################################
     # Original Formatting Splitting
     # These checks rely upon the original formatting. This is in order to
@@ -488,6 +487,8 @@ class FormatDecisionState(object):
     else:
       self._AddTokenOnCurrentLine(dry_run)
 
+    penalty += self._CalculateComprehensionState(newline)
+
     return self.MoveStateToNextToken() + penalty
 
   def _AddTokenOnCurrentLine(self, dry_run):
@@ -587,46 +588,6 @@ class FormatDecisionState(object):
 
     return penalty + 10
 
-  def _GetNewlineColumn(self):
-    """Return the new column on the newline."""
-    current = self.next_token
-    previous = current.previous_token
-    top_of_stack = self.stack[-1]
-
-    if current.spaces_required_before > 2 or self.line.disable:
-      return current.spaces_required_before
-
-    if current.OpensScope():
-      return top_of_stack.indent if self.paren_level else self.first_indent
-
-    if current.ClosesScope():
-      if (previous.OpensScope() or
-          (previous.is_comment and previous.previous_token is not None and
-           previous.previous_token.OpensScope())):
-        return max(0,
-                   top_of_stack.indent - style.Get('CONTINUATION_INDENT_WIDTH'))
-      return top_of_stack.closing_scope_indent
-
-    if (previous and previous.is_string and current.is_string and
-        format_token.Subtype.DICTIONARY_VALUE in current.subtypes):
-      return previous.column
-
-    if style.Get('INDENT_DICTIONARY_VALUE'):
-      if previous and (previous.value == ':' or previous.is_pseudo_paren):
-        if format_token.Subtype.DICTIONARY_VALUE in current.subtypes:
-          return top_of_stack.indent
-
-    if (_IsCompoundStatement(self.line.first) and
-        (not style.Get('DEDENT_CLOSING_BRACKETS') or
-         style.Get('SPLIT_BEFORE_FIRST_ARGUMENT'))):
-      token_indent = (
-          len(self.line.first.whitespace_prefix.split('\n')[-1]) +
-          style.Get('INDENT_WIDTH'))
-      if token_indent == top_of_stack.indent:
-        return top_of_stack.indent + style.Get('CONTINUATION_INDENT_WIDTH')
-
-    return top_of_stack.indent
-
   def MoveStateToNextToken(self):
     """Calculate format decision state information and move onto the next token.
 
@@ -683,6 +644,115 @@ class FormatDecisionState(object):
 
     return penalty
 
+  def _CalculateComprehensionState(self, newline):
+    """Makes required changes to comprehension state.
+
+    Args:
+      newline: Whether the current token is to be added on a newline.
+
+    Returns:
+      The penalty for the token-newline combination given the current
+      comprehension state.
+    """
+    current = self.next_token
+    previous = current.previous_token
+    top_of_stack = self.comp_stack[-1] if self.comp_stack else None
+    penalty = 0
+
+    if top_of_stack is not None:
+      # Check if the token terminates the current comprehension.
+      if current == top_of_stack.closing_bracket:
+        last = self.comp_stack.pop()
+        # Lightly penalize comprehensions that are split across multiple lines.
+        if last.has_interior_split:
+          penalty += style.Get('SPLIT_PENALTY_COMPREHENSION')
+
+        return penalty
+
+      if newline:
+        top_of_stack.has_interior_split = True
+
+    if (format_token.Subtype.COMP_EXPR in current.subtypes and
+        format_token.Subtype.COMP_EXPR not in previous.subtypes):
+      self.comp_stack.append(object_state.ComprehensionState(current))
+      return penalty
+
+    if (current.value == 'for' and
+        format_token.Subtype.COMP_FOR in current.subtypes):
+      if top_of_stack.for_token is not None:
+        # Treat nested comprehensions like normal comp_if expressions.
+        # Example:
+        #     my_comp = [
+        #         a.qux + b.qux
+        #         for a in foo
+        #   -->   for b in bar   <--
+        #         if a.zut + b.zut
+        #     ]
+        if (style.Get('SPLIT_COMPLEX_COMPREHENSION') and
+            top_of_stack.has_split_at_for != newline and
+            (top_of_stack.has_split_at_for or
+             not top_of_stack.HasTrivialExpr())):
+          penalty += split_penalty.UNBREAKABLE
+      else:
+        top_of_stack.for_token = current
+        top_of_stack.has_split_at_for = newline
+
+        # Try to keep trivial expressions on the same line as the comp_for.
+        if (style.Get('SPLIT_COMPLEX_COMPREHENSION') and newline and
+            top_of_stack.HasTrivialExpr()):
+          penalty += split_penalty.CONNECTED
+
+    if (format_token.Subtype.COMP_IF in current.subtypes and
+        format_token.Subtype.COMP_IF not in previous.subtypes):
+      # Penalize breaking at comp_if when it doesn't match the newline structure
+      # in the rest of the comprehension.
+      if (style.Get('SPLIT_COMPLEX_COMPREHENSION') and
+          top_of_stack.has_split_at_for != newline and
+          (top_of_stack.has_split_at_for or not top_of_stack.HasTrivialExpr())):
+        penalty += split_penalty.UNBREAKABLE
+
+    return penalty
+
+  def _GetNewlineColumn(self):
+    """Return the new column on the newline."""
+    current = self.next_token
+    previous = current.previous_token
+    top_of_stack = self.stack[-1]
+
+    if current.spaces_required_before > 2 or self.line.disable:
+      return current.spaces_required_before
+
+    if current.OpensScope():
+      return top_of_stack.indent if self.paren_level else self.first_indent
+
+    if current.ClosesScope():
+      if (previous.OpensScope() or
+          (previous.is_comment and previous.previous_token is not None and
+           previous.previous_token.OpensScope())):
+        return max(0,
+                   top_of_stack.indent - style.Get('CONTINUATION_INDENT_WIDTH'))
+      return top_of_stack.closing_scope_indent
+
+    if (previous and previous.is_string and current.is_string and
+        format_token.Subtype.DICTIONARY_VALUE in current.subtypes):
+      return previous.column
+
+    if style.Get('INDENT_DICTIONARY_VALUE'):
+      if previous and (previous.value == ':' or previous.is_pseudo_paren):
+        if format_token.Subtype.DICTIONARY_VALUE in current.subtypes:
+          return top_of_stack.indent
+
+    if (_IsCompoundStatement(self.line.first) and
+        (not style.Get('DEDENT_CLOSING_BRACKETS') or
+         style.Get('SPLIT_BEFORE_FIRST_ARGUMENT'))):
+      token_indent = (
+          len(self.line.first.whitespace_prefix.split('\n')[-1]) +
+          style.Get('INDENT_WIDTH'))
+      if token_indent == top_of_stack.indent:
+        return top_of_stack.indent + style.Get('CONTINUATION_INDENT_WIDTH')
+
+    return top_of_stack.indent
+
   def _FitsOnLine(self, start, end):
     """Determines if line between start and end can fit on the current line."""
     length = end.total_length - start.total_length
@@ -751,6 +821,20 @@ class FormatDecisionState(object):
     length += len(entry_start.value)
     return length + self.stack[-2].indent <= self.column_limit
 
+  def _ArgumentListHasDictionaryEntry(self, token):
+    """Check if the function argument list has a dictionary as an arg."""
+    if _IsArgumentToFunction(token):
+      while token:
+        if token.value == '{':
+          length = token.matching_bracket.total_length - token.total_length
+          return length + self.stack[-2].indent > self.column_limit
+        if token.ClosesScope():
+          break
+        if token.OpensScope():
+          token = token.matching_bracket
+        token = token.next_token
+    return False
+
 
 _COMPOUND_STMTS = frozenset(
     {'for', 'while', 'if', 'elif', 'with', 'except', 'def', 'class'})
@@ -773,12 +857,20 @@ def _IsFunctionCallWithArguments(token):
     if token.value == '(':
       token = token.next_token
       return token and token.value != ')'
-    elif token.name not in {'NAME', 'DOT'}:
+    elif token.name not in {'NAME', 'DOT', 'EQUAL'}:
       break
     token = token.next_token
   return False
 
 
+def _IsArgumentToFunction(token):
+  bracket = unwrapped_line.IsSurroundedByBrackets(token)
+  if not bracket or bracket.value != '(':
+    return False
+  previous = bracket.previous_token
+  return previous and previous.is_name
+
+
 def _GetLengthOfSubtype(token, subtype, exclude=None):
   current = token
   while (current.next_token and subtype in current.subtypes and
